/*++++++++++++++++++ TFG +++++++++++++++++++
Desarollado por Daniel Sierra
Para Robot auto balanceado
Estudiante de ingenieria electrónica y autómatica
Universidad de Valladolid
Tutor Luis Carlos Herrero
2018/2019

Bibliografia:

 +++++++++++++++++++++++++++++++++++++++++++++*/


// +++++++LIBRERIAS++++++++++++
#include "Wire.h"// Conexión de I2C
#include <digitalWriteFast.h>//Para leer de los encoders
#include <SoftwareSerial.h>// Para el uso del módulo bluetooth

// ++++++PINES++++++++++++++++

#define pinEncoderIzqA 3
#define pinEncoderIzqB 5
#define pinEncoderDerA 2
#define pinEncoderDerB 4
#define PWMIZQ 9
#define PWMDER 10
#define DIRDER 7
#define DIRIZQ 8
#define RX 11
#define TX 6


//+++++++CALCULO DE ÁNGULO ++++++++++++++
//Ratios de conversion
#define A_R 16384.0
#define G_R 131.0
//Conversion de radianes a grados 180/PI
#define RAD_A_DEG = 57.295779
//Frecuencia del integrador
#define FHz 55
float Acc[2];
float Gy[2];
float Angle[2];
float dt = 0.004;
unsigned long tMuestreo = 0;

//++++++++ENCODER++++++++++++++++++++++++++
#define interEncoderIzq 1 //Interrupción para encoder
#define encoderIzquierdoReversa //Para seleccionar el sentido de giro positivo
#define gradosPorPulso 0.50f //Grados por pulso del encoder
#define interEncoderDer 0 //Interrupción para encoder

volatile bool _lecturaEncoderIzqB;//Variable para lectura Encoder
volatile bool _lecturaEncoderDerB;//Variable para lectura Encoder
volatile long _pulsosEncoderIzq = 0;//Pasos Encoder
volatile long _pulsosEncoderDer = 0;//Pasos Encoder
static double posicionRueda, ultimaPosicionRueda, anguloRotacion;
static unsigned long int tiempoRotRueda;
static float velocidadRueda;

//+++++++++++BLUETOOTH+++++++++++++++++++++++
String mensaje;//Variable para almacenar el mensaje recibido
bool robotHabilitado = 1;//Habilita y desahabilita el robot
char separador = ',';//Caracter que que define la separación entre los datos
SoftwareSerial BT(TX, RX);// Define los pines RX,TX para la comunicación Bluetooth
String dato1, dato2, dato3;//Variables donde colocan los datos al dividir el mensaje

//++++++++++MPU6050++++++++++++++++++
//MPU-6050 da los valores en enteros de 16 bits
//Variables para almacenar los detos enviados por el módulo
int16_t AcX, AcY, AcZ, GyX, GyY, GyZ;
//Direccion I2C
#define MPU 0x68

//++++++++PID Angulo++++++++++++++++++
float Kp = 5;// Constante proporcional
float Ki = 3;// Constante integral
float Kd = 2.5;// Constante derivativa
double SetpointAngulo = -.7; // Referencia del PID
bool dir = 0; // Define el sentido de giro
double Output; // Salida del PID
float Angulo; // Ángulo medido
float Estado_integral = 0; // Suma de la parte proporcional
float Error, ultimoError; // Variables del error del controlador
float Kp_fb = 0, Ki_fb = 0, Kd_fb = 0; // Variables de la multipicación entre la const. con el error 


void CalculoAngulo() {

  //Leer los valores del Acelerometro de la MPU
  Wire.beginTransmission(MPU);
  Wire.write(0x3B); //Pedir el registro 0x3B - corresponde al AcX
  Wire.endTransmission(false);
  Wire.requestFrom(MPU, 6, true); //A partir del 0x3B, se piden 6 registros
  AcX = Wire.read() << 8 | Wire.read(); //Cada valor ocupa 2 registros
  AcY = Wire.read() << 8 | Wire.read();
  AcZ = Wire.read() << 8 | Wire.read();

  //A partir de los valores del acelerometro, se calculan los angulos Y, X
  //respectivamente, con la formula de la tangente.
  Acc[1] = atan(-1 * (AcX / A_R) / sqrt(pow((AcY / A_R), 2) + pow((AcZ / A_R), 2))) * RAD_TO_DEG;
  Acc[0] = atan((AcY / A_R) / sqrt(pow((AcX / A_R), 2) + pow((AcZ / A_R), 2))) * RAD_TO_DEG;

  //Leer los valores del Giroscopio
  Wire.beginTransmission(MPU);
  Wire.write(0x43); // Se solicita el registro 0x43
  Wire.endTransmission(false);
  Wire.requestFrom(MPU, 4, true); //A diferencia del Acelerometro, solo se piden 4 registros
  GyX = Wire.read() << 8 | Wire.read();
  GyY = Wire.read() << 8 | Wire.read();

  //Calculo del angulo del Giroscopio
  Gy[0] = GyX / G_R;
  Gy[1] = GyY / G_R;


  //Aplicar el Filtro Complementario
  Angle[1] = 0.98 * (Angle[1] + Gy[1] * dt) + 0.02 * Acc[1];
  //Se asigna el valor a la entrada del controlador
  Angulo = Angle[1] ;
}
//++++Calculo para generar la salida del sistema mediante PID
void CalculoPID() {
  // Si se esta por el rango de ángulo de funcionamiento
  if (abs(Angulo) < 20) {
    //Se almacena el valor del error para el caculo de integral
    ultimoError = Error;
    Error = SetpointAngulo - Angulo;
    Estado_integral = Estado_integral + Error / FHz;
    // Se límita el valor del Ingegral
    Estado_integral = constrain(Estado_integral, -30.0f, 30.0f);
    Kp_fb = Kp * Error;
    Ki_fb = Ki * Estado_integral;
    Kd_fb = -Kd * Gy[0];
    Output = Kp_fb + Ki_fb + Kd_fb;
    // Se limita el valor de la salida
    Output = constrain(Output, -250, 250);
    //Se define el sentido de giro
    if (Output <= 0) {
      dir = 0;
    }
    else {
      dir = 1;
    }
  }
}

//++++Función para detección de pulsos en la rueda izquierda
void fInterIzq()
{
  // lee en pin del otro encoder para saber el sentido
  _lecturaEncoderIzqB = digitalReadFast(pinEncoderIzqB);  
  // conteo depemdiendo del sentido de giro que lleva el robot
  //hacia adelante
#ifdef encoderIzquierdoReversa
  _pulsosEncoderIzq -= _lecturaEncoderIzqB ? -1 : +1;
#else
  _pulsosEncoderIzq += _lecturaEncoderIzqB ? -1 : +1;
#endif
}

//++++Función para detección de pulsos en la rueda izquierda
void fInterDer()
{
  // lee en pin del otro encoder para saber el sentido
  _lecturaEncoderDerB = digitalReadFast(pinEncoderDerB);   // lee en pin de entrada

  // conteo depemdiendo del sentido de giro que lleva el robot
  //hacia adelante
#ifdef encoderDerechaReversa
  _pulsosEncoderDer -= _lecturaEncoderDerB ? -1 : +1;
#else
  _pulsosEncoderDer += _lecturaEncoderDerB ? -1 : +1;
#endif
}

//+++Para el calculo de la velocidad a partir del conteo de pulsos
void CalculoVelocidad() {
  //Tiene un tiempo de muestreo de 0.01s
  if (millis() - tiempoRotRueda > 10) {
    tiempoRotRueda = millis();
    posicionRueda = 0.5f * (_pulsosEncoderDer + _pulsosEncoderIzq) * gradosPorPulso;
    velocidadRueda = 10.0f * (posicionRueda - ultimaPosicionRueda);
    anguloRotacion = 0.5f * (_pulsosEncoderIzq - _pulsosEncoderDer) * gradosPorPulso;
    ultimaPosicionRueda = posicionRueda;
  }
}

//++++INICIALIZACIÓN DE MÓDULO BLUETOOTH HC-05
void InicializacionBT() {
  //Declara la velocidad de comunicacón con el Bluetooth
  BT.begin(9600);
}

//++++INICIALIZACIÓN DEL COMUNICACIÓN I2C PARA EL MPU-5060
void InicializacionI2C() {
  //Se comienza la conexión
  Wire.begin();
  //Determina la vel. de reloj de transmición a 400KHz
  Wire.setClock(400000UL);
  Wire.beginTransmission(MPU);
  Wire.write(0x6B);
  //Inicializa el módulo
  Wire.write(0);
  Wire.endTransmission(true);

}

//+++++++++++RECEPCIÓN DE INFORMACIÓN
void RecepcionInformacion() {
  //Siempre que se tenga datos a la espera de ser leidos
  if (BT.available()) {
    //Divide el mensaje para ponerlo en las constantes
    String dato = BT.readStringUntil('\n');
    dato1 = dato.substring(0, dato.indexOf(separador));
    dato = dato.substring(dato1.length() + 1);
    dato2 = dato.substring(0, dato.indexOf(separador));
    dato = dato.substring(dato2.length() + 1);
    dato3 = dato.substring(0, dato.indexOf(separador));
    Ki = dato1.toFloat();
    Kp = dato2.toFloat();
    Kd = dato3.toFloat();
    //Si el mensaje recibido es "-1,-1,-1" se desactiva
    if(Kp == -1 && Ki == -1 && Kd == -1){
      robotHabilitado = 0;
    }
    else{
      robotHabilitado = 1;
    }

  }
}

//+++++++ ESCRIBIR EN LA SALIDA
void EscribirSenalControl(){
    if(robotHabilitado && abs(Angulo) <  20){
      digitalWrite(DIRIZQ, dir);
      digitalWrite(DIRDER, dir);
      analogWrite(PWMIZQ, abs(Output));
      analogWrite(PWMDER, abs(Output));
    }

  else {
    analogWrite(PWMIZQ, abs(0));
    analogWrite(PWMDER, abs(0));
    RecepcionInformacion();
  }
}

//++++Toma de valores para estabilizar muestreo del ángulo
void EstabilizarAngulo(){
  int contador = 0;
  while(contador < 100){
    CalculoAngulo();
    contador ++;
    delay(5);
  }
}
//++++Se crean las interrupciones para los Encoder
void InicializacionInt(){
  attachInterrupt(interEncoderIzq, fInterIzq, RISING);
  attachInterrupt(interEncoderDer, fInterDer, RISING);
}

//++++Inicialización del robot
void setup()
{
  InicializacionI2C();
  InicializacionBT();
  InicializacionInt();
  EstabilizarAngulo();
}

//++++Bucle general
void loop()
{
  if (micros() - tMuestreo > 900) {
    tMuestreo = micros();
    CalculoAngulo();
    CalculoPID();
    EscribirSenalControl();
    CalculoVelocidad();
  }

}
